'*************************************************************************
'Contents of this file are (C)Copyright 2009 by Circuit Check Inc.  Any use of
'the contents of this file without an explicit license from Circuit Check Inc
'is THEFT.  Copyright violators will be prosecuted.
'
'        File : deviceclass.vb
'        Type : Class
'    Language : Visual Basic
'    Revision : 1.0
'      Author : M. van Meeteren
'        Date : 022009
' Description : Base class for connected equipment
'
'*************************************************************************
Imports System.Threading
Imports System.IO

Public Enum Device_Type
  dtNone
  dtDMM
  dtDIO
End Enum

Friend MustInherit Class Device_Class
  Friend InittedOK As Boolean = False

#Region "Status/Error Events"
  Friend Equip_Error As String = String.Empty

  Friend Event DeviceStatusEvent(Status As String)
  Protected Sub DeviceStatus(Status As String, Optional ForceRaise As Boolean = False)
    If Not ForceRaise Then Exit Sub
    Status = Strip(Status)
    RaiseEvent DeviceStatusEvent(Status)
  End Sub
  Protected Sub DeviceError(sError As String)
    Equip_Error = Strip(sError)
    RaiseEvent DeviceStatusEvent(MyDevice & ": " & Equip_Error)
  End Sub
  Protected Sub DeviceException(sError As String)
    DeviceError(sError)
  End Sub
  Protected Function CheckError(bError As Boolean) As Boolean
    If bError Then DeviceError(Equip_Error)
    Return Not bError
  End Function
#End Region

#Region "Initialize/Close"
  Private Function ReadSlotAddress(DeviceSection As String, _
                                   INIFileName As String) As Boolean
    Dim INIParam As String = String.Empty

    ReadSlotAddress = False
    INIParam = ReadINIFile(DeviceSection, "cardslot", String.Empty, INIFileName)
    ' error is neither is set
    If (INIParam = String.Empty) Then
      DeviceError("slot location not found")
      Exit Function
    End If
    Try
      CardSlot = Integer.Parse(INIParam)
      MyAddress = CardSlot * IsHostedCard
      Return True
    Catch ex As Exception
      DeviceError("slot location parse error: " + ex.Message)
      Exit Function
    End Try
  End Function

  Friend Function ReadDeviceAndAddress(DeviceSection As String, _
                                       INIFileName As String) As Boolean
    Dim INIParam As String = String.Empty
    Dim INIParam2 As String = String.Empty
    Dim ChannelKey As String = "basechannel"

    ReadDeviceAndAddress = False
    If IsHostedCard > 0 Then
      If Not ReadSlotAddress(DeviceSection, INIFileName) Then Return False
    Else
      INIParam = ReadINIFile(DeviceSection, ChannelKey, String.Empty, INIFileName)
      If (INIParam = String.Empty) Then
        DeviceError(ChannelKey & " not found")
        Exit Function
      End If
      Try
        MyAddress = Integer.Parse(INIParam)
      Catch ex As Exception
        DeviceError(ChannelKey & " parse error: " + ex.Message)
        Exit Function
      End Try
    End If

    ' hosted cards are tied to a slot, so don't read the device
    If (IsHostedCard = 0) Then
      INIParam = ReadINIFile(DeviceSection, "device", String.Empty, INIFileName)
      If (INIParam = String.Empty) Then
        DeviceError("Device Parameter Not Specified")
        Exit Function
      End If
      MyDevice = INIParam.ToUpper.Trim
    End If

    ReadDeviceAndAddress = True
  End Function

  Friend Overridable Sub CloseDevice()
    ' inherit this to dispose things like COM ports, shut off DC supplies, etc
  End Sub
#End Region

#Region "Channel Numbers"
  Protected MyAddress As Integer = 1000
  Protected MyDevice As String = String.Empty

  Private IsHostedCard As Integer = 0
  Protected CardSlot As Integer = 0
  Protected Sub HostedCard(CardAddressMultiplier As Integer)
    IsHostedCard = CardAddressMultiplier
  End Sub

  Friend Class Device_Address_Range
    Public RangeLow As Integer
    Public RangeHigh As Integer
    Public DeviceDescription As String
  End Class

  Friend MustOverride Function DeviceType() As Device_Type

  Friend MustOverride Function TextChannelList() As String()

#End Region

#Region "Read Instrument Error"
  Protected Overridable Sub ErrorQuery(ByRef errCode As Integer, ByRef errMsg As String)
    errCode = 0
    errMsg = "OK"
  End Sub

  Private Function AllNumbers(Code As String) As Boolean
    Dim AllNumber As Boolean = True

    ' check to see if it's JUST a number
    While AllNumber AndAlso Code.Length > 0
      If "01234567890+-".IndexOf(Code(0)) < 0 Then
        Return False
      Else
        Code = Code.Substring(1)
      End If
    End While
    Return True
  End Function

  Protected Sub ParseQuery(QueryResult As String, _
                           ByRef errCode As Integer, _
                           ByRef errMsg As String, _
                           Optional NoErrorString As String = "NO ERROR")
    Dim Pos As Integer
    Dim AllNumber As Boolean = True
    Dim sCode As String

    QueryResult = QueryResult.Trim

    ' default returns in case no code is part of the string
    errCode = -1
    errMsg = QueryResult

    ' try to parse out the error number from the string
    Pos = QueryResult.IndexOf(","c)
    If Pos > 0 Then
      sCode = errMsg.Substring(0, Pos)
      AllNumber = AllNumbers(sCode)
      If AllNumber Then _
        errMsg = errMsg.Substring(Pos + 1)
    Else
      sCode = errMsg
      AllNumber = AllNumbers(sCode)
      If AllNumber Then _
        errMsg = String.Empty
    End If

    ' we have a good code!
    If AllNumber Then
      Try
        errCode = Integer.Parse(sCode)
        If errMsg = String.Empty Then
          ' instrument gave us just a numerical value
          errMsg = SCPIError.SCPIErrorToString(errCode)
        Else
          ' instrument gave us additional info after the comma
          ' errMsg = errMsg.Substring(Pos + 1)
          errMsg = errMsg.Trim
        End If
      Catch ex As Exception
        ' this SHOULD never happen, we already prequalified sCode
        errCode = -999
        errMsg = "exception parsing <" & sCode & ">: " & errMsg
      End Try

    Else ' not allnumber
      errCode = -1
      ' might be a message with a comma in it, or a message without a number
      If errMsg.ToUpper.Contains(NoErrorString) Then errCode = 0
    End If

  End Sub

  Private Function intReadInstrumentErrors(ByRef ErrorCodes As String, _
                                           AllErrors As Boolean, _
                                           ReportAnyError As Boolean) As Boolean
    ' Read instrument error queue until its empty.
    Dim errCode As Integer = 0
    Dim errMsg As String = String.Empty
    Dim GotError As Boolean = False

    ' if we hit an exception, get all errors
    ErrorCodes = String.Empty
    intReadInstrumentErrors = True

    Do
      ' give instrument three times to reply to error query before declaring it dead
      Dim Retries As Integer = 2
      Do
        Try
          ' query the instrument for the error
          ErrorQuery(errCode, errMsg)
          Exit Do
        Catch ex As Exception
          Dim Timeout As Boolean = _
            ex.Message.ToUpper.Contains("ERROR_TMO") OrElse _
            ex.Message.ToUpper.Contains("TIMEOUT")
          Retries -= 1
          ' see if we need to exit
          If (Retries < 0) Or (Not Timeout) Then
            If Timeout Then
              ' timeout, but retries expires
              ErrorCodes = "-365,ErrorQuery timeout (retried 3 times): " & ex.Message
            Else
              ' not a timeout, immediately bail
              ErrorCodes = "-999,ErrorQuery exception: " & ex.Message
            End If
            InittedOK = False
            Exit Function
          End If
        End Try
      Loop

      ' append the result to the existing codes, where we're reading all of them,
      ' or 0,NO ERROR when we're only reading 1
      If (errCode <> 0) Or (ErrorCodes = String.Empty) Then
        If (ErrorCodes <> String.Empty) Then ErrorCodes += " | "
        ErrorCodes += errCode & "," & errMsg
      End If

      ' check to see if we got any non-zero codes so far
      GotError = GotError Or (errCode <> 0)

      If AllErrors Then
        ' if we're reading all the errors, loop until we get a 0, 
        If (errCode = 0) Then Exit Do
      Else
        ' otherwise return what we got
        Exit Do
      End If
    Loop

    intReadInstrumentErrors = GotError And ReportAnyError
  End Function

  ' reads first error in queue, true if errorcode is not zero
  Protected Function ReadInstrumentError(ByRef ErrorCodes As String) As Boolean
    Return intReadInstrumentErrors(ErrorCodes, False, True)
  End Function

  ' reads all errors in queue, returns true if any are found
  Protected Function ReadInstrumentErrors(ByRef ErrorCodes As String) As Boolean
    Return intReadInstrumentErrors(ErrorCodes, True, True)
  End Function

  ' reads all errors, returns false (no error) unless an exception happens
  Protected Function FlushInstrumentErrors(ByRef ErrorCodes As String) As Boolean
    Return intReadInstrumentErrors(ErrorCodes, True, False)
  End Function

#End Region

#Region "Shared utility functions"
  Friend Shared Function Strip(S As String) As String
    Dim Index As Integer = 0

    While (Index < S.Length)
      If (S(Index) < " "c) Or (S(Index) > "~"c) Then
        S = S.Remove(Index, 1)
      Else
        Index += 1
      End If
    End While
    Return S
  End Function
#End Region

End Class

